#include <jni.h>
#include <jvmti.h>
#include <stdlib.h>
#include <string.h>

#import "utils.h"

JNIEnv * getJniEnv(JavaVM * javaVM) {
	JNIEnv * ret;
	const jint envResult = (*javaVM)->GetEnv(javaVM, (void**) &ret,
			JNI_VERSION_1_6);
	if (envResult == JNI_EDETACHED) {
		JavaVMAttachArgs args;
		args.version = JNI_VERSION_1_6;
		const jint result = (*javaVM)->AttachCurrentThreadAsDaemon(javaVM,
				(void**) &ret, &args);
		if (result != 0) {
			fprintf(stderr, "ERROR: Unable to attach, error=%d\n", result);
			abort();
		}
	}
	return ret;
}

void set_caps(jvmtiEnv * jvmti) {
	jvmtiCapabilities capsPotential;
	const jvmtiError errPotential = (*jvmti)->GetPotentialCapabilities(jvmti,
			&capsPotential);
	CHECK_JVMTI_ERROR(jvmti, errPotential);
	if (capsPotential.can_tag_objects != 1) {
		fprintf(stderr,
				"ERROR: VM cannot give us capability to tag objects.\n");
		abort();
	}
	fprintf(stderr,
					"OK: VM can give us capability to tag objects.\n");

	jvmtiCapabilities caps;
	(*jvmti)->GetCapabilities(jvmti, &caps);
	caps.can_tag_objects = 1;
	caps.can_generate_object_free_events = 1;
	const jvmtiError err = (*jvmti)->AddCapabilities(jvmti, &caps);
	CHECK_JVMTI_ERROR(jvmti, err);
}

void set_callbacks(jvmtiEnv * jvmti, void (*gc_callback)(jvmtiEnv *, jlong)) {
	jvmtiEventCallbacks callbacks;
	(void) memset(&callbacks, 0, sizeof(callbacks));
	callbacks.ObjectFree = gc_callback;
	const jvmtiError err = (*jvmti)->SetEventCallbacks(jvmti, &callbacks,
			(jint) sizeof(callbacks));
	CHECK_JVMTI_ERROR(jvmti, err);

	const jvmtiError enable_err = (*jvmti)->SetEventNotificationMode(jvmti,
			JVMTI_ENABLE, JVMTI_EVENT_OBJECT_FREE, (jthread) NULL);
	CHECK_JVMTI_ERROR(jvmti, enable_err);
}

static void deallocate(jvmtiEnv *jvmti, void *p) {
	jvmtiError err;

	err = (*jvmti)->Deallocate(jvmti, (unsigned char *) p);
	if (err != JVMTI_ERROR_NONE) {
		fprintf(stderr, "ERROR: JVMTI Deallocate error err=%d\n", err);
		abort();
	}
}

static char * getErrorName(jvmtiEnv *jvmti, jvmtiError errnum) {
	jvmtiError err;
	char *name;

	err = (*jvmti)->GetErrorName(jvmti, errnum, &name);
	if (err != JVMTI_ERROR_NONE) {
		fprintf(stderr, "ERROR: JVMTI GetErrorName error err=%d\n", err);
		abort();
	}
	return name;
}

void checkJvmtiError(jvmtiEnv *jvmti, jvmtiError err, char *file, int line) {
	if (err != JVMTI_ERROR_NONE) {
		char *name;

		name = getErrorName(jvmti, err);
		fprintf(stderr, "ERROR: JVMTI error err=%d(%s) in %s:%d\n", err, name,
				file, line);
		deallocate(jvmti, name);
		abort();
	}
}
