package ch.usi.inf.sape.ld;

import static org.objectweb.asm.Opcodes.*;

import java.util.List;

import org.objectweb.asm.Handle;
import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.IntInsnNode;
import org.objectweb.asm.tree.InvokeDynamicInsnNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.MultiANewArrayInsnNode;
import org.objectweb.asm.tree.TypeInsnNode;
import org.objectweb.asm.tree.analysis.AnalyzerException;
import org.objectweb.asm.tree.analysis.Interpreter;

public class LeakDetectorInterpreter extends Interpreter<LeakDetectorValue> {

  private final ClassNode cn;

  private final MethodNode mn;

  protected LeakDetectorInterpreter(final int api, final ClassNode cn,
      final MethodNode mn) {
    super(api);
    this.cn = cn;
    this.mn = mn;
  }

  @Override
  public LeakDetectorValue newValue(final Type type) {
    if (type == null) {
      return new LeakDetectorValue(null);
    }
    switch (type.getSort()) {
    case Type.VOID:
      return null;
    case Type.BOOLEAN:
    case Type.CHAR:
    case Type.BYTE:
    case Type.SHORT:
    case Type.INT:
      return new LeakDetectorValue(Type.INT_TYPE);
    case Type.FLOAT:
      return new LeakDetectorValue(Type.FLOAT_TYPE);
    case Type.LONG:
      return new LeakDetectorValue(Type.LONG_TYPE);
    case Type.DOUBLE:
      return new LeakDetectorValue(Type.DOUBLE_TYPE);
    case Type.ARRAY:
    case Type.OBJECT:
      return new LeakDetectorValue(Type.getObjectType("java/lang/Object"));
    default:
      throw new Error("Internal error");
    }

  }

  @Override
  public LeakDetectorValue newOperation(final AbstractInsnNode insn)
      throws AnalyzerException {
    switch (insn.getOpcode()) {
    case ACONST_NULL:
      return new LeakDetectorValue(Type.getObjectType("null"));
    case ICONST_M1:
    case ICONST_0:
    case ICONST_1:
    case ICONST_2:
    case ICONST_3:
    case ICONST_4:
    case ICONST_5:
    case BIPUSH:
    case SIPUSH:
      return new LeakDetectorValue(Type.INT_TYPE);
    case LCONST_0:
    case LCONST_1:
      return new LeakDetectorValue(Type.LONG_TYPE);
    case FCONST_0:
    case FCONST_1:
    case FCONST_2:
      return new LeakDetectorValue(Type.FLOAT_TYPE);
    case DCONST_0:
    case DCONST_1:
      return new LeakDetectorValue(Type.DOUBLE_TYPE);
    case LDC:
      final Object cst = ((LdcInsnNode) insn).cst;
      if (cst instanceof Integer) {
        return new LeakDetectorValue(Type.INT_TYPE);
      } else if (cst instanceof Float) {
        return new LeakDetectorValue(Type.FLOAT_TYPE);
      } else if (cst instanceof Long) {
        return new LeakDetectorValue(Type.LONG_TYPE);
      } else if (cst instanceof Double) {
        return new LeakDetectorValue(Type.DOUBLE_TYPE);
      } else if (cst instanceof String) {
        return new LeakDetectorValue(Type.getObjectType("java/lang/String"));
      } else if (cst instanceof Type) {
        final int sort = ((Type) cst).getSort();
        if (sort == Type.OBJECT || sort == Type.ARRAY) {
          return new LeakDetectorValue(Type.getObjectType("java/lang/Class"));
        } else if (sort == Type.METHOD) {
          return new LeakDetectorValue(
              Type.getObjectType("java/lang/invoke/MethodType"));
        } else {
          throw new IllegalArgumentException("Illegal LDC constant " + cst);
        }
      } else if (cst instanceof Handle) {
        return new LeakDetectorValue(
            Type.getObjectType("java/lang/invoke/MethodHandle"));
      } else {
        throw new IllegalArgumentException("Illegal LDC constant " + cst);
      }
    case JSR:
      return new LeakDetectorValue(Type.VOID_TYPE);
    case GETSTATIC:
      return new LeakDetectorValue(Type.getType(((FieldInsnNode) insn).desc));
    case NEW:
      final LeakDetectorValue ret = new LeakDetectorValue(
          Type.getObjectType(((TypeInsnNode) insn).desc));
      ret.getProvenance().add(new LeakDetectorProvenance(insn));
      return ret;
    default:
      throw new Error("Internal error.");
    }
  }

  @Override
  public LeakDetectorValue copyOperation(final AbstractInsnNode insn,
      final LeakDetectorValue value) throws AnalyzerException {
    return value;
  }

  @Override
  public LeakDetectorValue unaryOperation(final AbstractInsnNode insn,
      final LeakDetectorValue value) throws AnalyzerException {
    switch (insn.getOpcode()) {
    case INEG:
    case IINC:
    case L2I:
    case F2I:
    case D2I:
    case I2B:
    case I2C:
    case I2S:
      return new LeakDetectorValue(Type.INT_TYPE);
    case FNEG:
    case I2F:
    case L2F:
    case D2F:
      return new LeakDetectorValue(Type.FLOAT_TYPE);
    case LNEG:
    case I2L:
    case F2L:
    case D2L:
      return new LeakDetectorValue(Type.LONG_TYPE);
    case DNEG:
    case I2D:
    case L2D:
    case F2D:
      return new LeakDetectorValue(Type.DOUBLE_TYPE);
    case IFEQ:
    case IFNE:
    case IFLT:
    case IFGE:
    case IFGT:
    case IFLE:
    case TABLESWITCH:
    case LOOKUPSWITCH:
    case IRETURN:
    case LRETURN:
    case FRETURN:
    case DRETURN:
    case ARETURN:
    case PUTSTATIC:
      return null;
    case GETFIELD:
      return new LeakDetectorValue(Type.getType(((FieldInsnNode) insn).desc));
    case NEWARRAY:
      switch (((IntInsnNode) insn).operand) {
      case T_BOOLEAN:
        return new LeakDetectorValue(Type.getType("[Z"));
      case T_CHAR:
        return new LeakDetectorValue(Type.getType("[C"));
      case T_BYTE:
        return new LeakDetectorValue(Type.getType("[B"));
      case T_SHORT:
        return new LeakDetectorValue(Type.getType("[S"));
      case T_INT:
        return new LeakDetectorValue(Type.getType("[I"));
      case T_FLOAT:
        return new LeakDetectorValue(Type.getType("[F"));
      case T_DOUBLE:
        return new LeakDetectorValue(Type.getType("[D"));
      case T_LONG:
        return new LeakDetectorValue(Type.getType("[J"));
      default:
        throw new AnalyzerException(insn, "Invalid array type");
      }
    case ANEWARRAY:
      String desc = ((TypeInsnNode) insn).desc;
      return new LeakDetectorValue(Type.getType("[" + Type.getObjectType(desc)));
    case ARRAYLENGTH:
      return new LeakDetectorValue(Type.INT_TYPE);
    case ATHROW:
      return null;
    case CHECKCAST:
      desc = ((TypeInsnNode) insn).desc;
      return new LeakDetectorValue(Type.getObjectType(desc));
    case INSTANCEOF:
      return new LeakDetectorValue(Type.INT_TYPE);
    case MONITORENTER:
    case MONITOREXIT:
    case IFNULL:
    case IFNONNULL:
      return null;
    default:
      throw new Error("Internal error.");
    }

  }

  @Override
  public LeakDetectorValue binaryOperation(final AbstractInsnNode insn,
      final LeakDetectorValue value1, final LeakDetectorValue value2)
      throws AnalyzerException {
    switch (insn.getOpcode()) {
    case IALOAD:
    case BALOAD:
    case CALOAD:
    case SALOAD:
    case IADD:
    case ISUB:
    case IMUL:
    case IDIV:
    case IREM:
    case ISHL:
    case ISHR:
    case IUSHR:
    case IAND:
    case IOR:
    case IXOR:
      return new LeakDetectorValue(Type.INT_TYPE);
    case FALOAD:
    case FADD:
    case FSUB:
    case FMUL:
    case FDIV:
    case FREM:
      return new LeakDetectorValue(Type.FLOAT_TYPE);
    case LALOAD:
    case LADD:
    case LSUB:
    case LMUL:
    case LDIV:
    case LREM:
    case LSHL:
    case LSHR:
    case LUSHR:
    case LAND:
    case LOR:
    case LXOR:
      return new LeakDetectorValue(Type.LONG_TYPE);
    case DALOAD:
    case DADD:
    case DSUB:
    case DMUL:
    case DDIV:
    case DREM:
      return new LeakDetectorValue(Type.DOUBLE_TYPE);
    case AALOAD:
      return new LeakDetectorValue(Type.getObjectType("java/lang/Object"));
    case LCMP:
    case FCMPL:
    case FCMPG:
    case DCMPL:
    case DCMPG:
      return new LeakDetectorValue(Type.INT_TYPE);
    case IF_ICMPEQ:
    case IF_ICMPNE:
    case IF_ICMPLT:
    case IF_ICMPGE:
    case IF_ICMPGT:
    case IF_ICMPLE:
    case IF_ACMPEQ:
    case IF_ACMPNE:
    case PUTFIELD:
      return null;
    default:
      throw new Error("Internal error.");
    }

  }

  @Override
  public LeakDetectorValue ternaryOperation(final AbstractInsnNode insn,
      final LeakDetectorValue value1, final LeakDetectorValue value2,
      final LeakDetectorValue value3) throws AnalyzerException {
    return null;
  }

  @Override
  public LeakDetectorValue naryOperation(final AbstractInsnNode insn,
      final List<? extends LeakDetectorValue> values) throws AnalyzerException {
    final int opcode = insn.getOpcode();

    if (opcode == MULTIANEWARRAY) {
      return new LeakDetectorValue(
          Type.getType(((MultiANewArrayInsnNode) insn).desc));
    }

    if (opcode == INVOKEDYNAMIC) {
      return new LeakDetectorValue(
          Type.getReturnType(((InvokeDynamicInsnNode) insn).desc));
    }

    if (opcode == INVOKESPECIAL) {
      final MethodInsnNode m = (MethodInsnNode) insn;
      if ("<init>".equals(m.name)) {
        final LeakDetectorValue v = values.get(0);
        for (final LeakDetectorProvenance provenance : v.getProvenance()) {
          if (provenance.isInitialized()) {
            throw new IllegalArgumentException(
                "Re-initializing already initialized value (" + m.owner + " "
                    + m.name + " " + m.desc + ") in " + cn.name + " " + mn.name
                    + " " + mn.desc + ": " + v.getType().toString() + " have "
                    + v.getProvenance().size() + " provenances");
          }
          provenance.setInitialized(true);
        }
        v.setInitializer(insn);
        return new LeakDetectorValue(
            Type.getReturnType(((MethodInsnNode) insn).desc));
      }
    }
    return new LeakDetectorValue(
        Type.getReturnType(((MethodInsnNode) insn).desc));
  }

  @Override
  public void returnOperation(final AbstractInsnNode insn,
      final LeakDetectorValue value, final LeakDetectorValue expected)
      throws AnalyzerException {
  }

  @Override
  public LeakDetectorValue merge(final LeakDetectorValue v,
      final LeakDetectorValue w) {
    // System.out.println("merging " + v + " and " + w);
    return v.merge(w);
  }

}
