package ch.usi.inf.sape.ld;

import static ch.usi.inf.sape.ld.LeakDetectorUtils.*;
import static org.objectweb.asm.Opcodes.*;

import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.security.ProtectionDomain;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.analysis.Analyzer;
import org.objectweb.asm.tree.analysis.AnalyzerException;
import org.objectweb.asm.tree.analysis.Frame;

public class LeakDetectorTransformer implements ClassFileTransformer {

  private static final String INIT_METHOD = "<init>";

  @Override
  public byte[] transform(final ClassLoader loader, final String className,
      final Class<?> classBeingRedefined,
      final ProtectionDomain protectionDomain, final byte[] classfileBuffer)
      throws IllegalClassFormatException {
    if (isInstrumentationClass(className) || isSdk(className)) {
      return classfileBuffer;
    }

    final ClassReader cr = new ClassReader(removeJsrRets(className,
        classfileBuffer));
    // final ClassReader cr = new ClassReader(classfileBuffer);
    final ClassNode cn = new ClassNode();
    cr.accept(cn, ASM4);

    for (final MethodNode mn : cn.methods) {
      final Analyzer<LeakDetectorValue> a = new Analyzer<LeakDetectorValue>(
          new LeakDetectorInterpreter(ASM4, cn, mn));
      try {
        final Frame<LeakDetectorValue>[] frames = a.analyze(cn.name, mn);

        final InsnList ilst = mn.instructions;

        int index = mn.maxLocals;
        int insnIndex = -1;
        for (final AbstractInsnNode insn : ilst.toArray()) {
          insnIndex++;
          switch (insn.getOpcode()) {
          case INVOKESPECIAL:
            final MethodInsnNode methodInsn = (MethodInsnNode) insn;
            if (methodInsn.name.equals(INIT_METHOD)) {
              final Type[] argumentTypes = Type
                  .getArgumentTypes(methodInsn.desc);
              final LeakDetectorValue v = frames[insnIndex - 1]
                  .getStack(argumentTypes.length);
              if (v != null && methodInsn == v.getInitializer()) {
                final LeakDetectorPair res = getInvokeCode(mn, index,
                    methodInsn.desc, false);
                mn.instructions.insertBefore(methodInsn, res.getInsns());
                index += res.getOffset();
                mn.instructions.insert(methodInsn, getInitCode());
              }
            }
            break;
          case INVOKEVIRTUAL:
          case INVOKEINTERFACE:
            final MethodInsnNode invoke = (MethodInsnNode) insn;
            if (!invoke.owner.equals(cn.name) || !mn.name.equals(INIT_METHOD)) {
              index++;
              final LeakDetectorPair res = getInvokeCode(mn, index,
                  invoke.desc, true);
              mn.instructions.insertBefore(invoke, res.getInsns());
              index += res.getOffset();
            }
            break;
          case GETFIELD:
            final FieldInsnNode getField = (FieldInsnNode) insn;
            if (!getField.owner.equals(cn.name) || !mn.name.equals(INIT_METHOD)) {
              mn.instructions.insertBefore(getField, getGetfieldCode());
            }
            break;
          case PUTFIELD:
            final FieldInsnNode putfield = (FieldInsnNode) insn;
            if (!putfield.owner.equals(cn.name) || !mn.name.equals(INIT_METHOD)) {
              index++;
              mn.instructions.insertBefore(putfield,
                  getPutfieldCode(mn, index, putfield.desc));
            }
            break;
          default:
            // this is OK
          }
        }
      } catch (final AnalyzerException e) {
        e.printStackTrace();
      }
    }

    final byte[] ret = writeBytecode(cn);
    // saveToFile(className, ret);
    return ret;
  }
}
