package ch.usi.inf.sape.ld;

import static org.objectweb.asm.Opcodes.*;

import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Type;
import org.objectweb.asm.commons.JSRInlinerAdapter;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.InsnNode;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LocalVariableNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.VarInsnNode;

public final class LeakDetectorUtils {

  private LeakDetectorUtils() {

  }

  private static final String LD = LeakDetector.class.getCanonicalName()
      .replace('.', '/');

  public static InsnList getInitCode() {
    final InsnList ret = new InsnList();
    ret.add(new MethodInsnNode(INVOKESTATIC, LD, "recordCreation",
        "(Ljava/lang/Object;)V"));
    return ret;
  }

  public static InsnList getPutfieldCode(final MethodNode mn, final int index,
      final String description) {
    final InsnList ret = new InsnList();
    final LabelNode start = new LabelNode();
    ret.add(start);
    ret.add(new VarInsnNode(getStore(description), index));
    ret.add(new InsnNode(DUP));
    ret.add(new MethodInsnNode(INVOKESTATIC, LD, "recordAccess",
        "(Ljava/lang/Object;)V"));
    ret.add(new VarInsnNode(getLoad(description), index));
    final LabelNode end = new LabelNode();
    ret.add(end);

    mn.localVariables.add(new LocalVariableNode("var" + index, description,
        null, start, end, index));
    return ret;
  }

  public static InsnList getGetfieldCode() {
    final InsnList ret = new InsnList();
    ret.add(new InsnNode(DUP));
    ret.add(new MethodInsnNode(INVOKESTATIC, LD, "recordAccess",
        "(Ljava/lang/Object;)V"));
    return ret;
  }

  public static LeakDetectorPair getInvokeCode(final MethodNode mn,
      final int freeIndex, final String description,
      final boolean stealImmediately) {
    final InsnList ret = new InsnList();
    final List<Type> argumentTypes = Arrays.asList(Type
        .getArgumentTypes(description));
    final List<Type> revArgTypes = Arrays.asList(Type
        .getArgumentTypes(description));
    Collections.reverse(revArgTypes);

    int offset = 0;
    final LabelNode start = new LabelNode();
    if (!argumentTypes.isEmpty()) {
      ret.add(start);
      for (final Type argumentType : revArgTypes) {
        offset++;
        ret.add(new VarInsnNode(getStore(argumentType.getDescriptor()),
            freeIndex + offset * 2));
      }
    }

    ret.add(new InsnNode(DUP));
    if (stealImmediately) {
      ret.add(new MethodInsnNode(INVOKESTATIC, LD, "recordAccess",
          "(Ljava/lang/Object;)V"));
    }

    if (!argumentTypes.isEmpty()) {
      for (final Type argumentType : argumentTypes) {
        ret.add(new VarInsnNode(getLoad(argumentType.getDescriptor()),
            freeIndex + offset * 2));
        offset--;
      }
      assert offset == 0;
      final LabelNode end = new LabelNode();
      ret.add(end);

      for (final Type argumentType : argumentTypes) {
        offset++;
        // TODO matej Uniquer names
        mn.localVariables.add(new LocalVariableNode("store" + offset * 2,
            argumentType.getDescriptor(), null, start, end, freeIndex + offset
                * 2));
      }
    }

    return new LeakDetectorPair(ret, offset * 2);
  }

  public static int getStore(final String desc) {
    if (desc.startsWith("[") || desc.startsWith("L")) {
      return ASTORE;
    }
    if (desc.startsWith("I") || desc.startsWith("S") || desc.startsWith("Z")
        || desc.startsWith("B") || desc.startsWith("C")) {
      return ISTORE;
    }
    if (desc.startsWith("F")) {
      return FSTORE;
    }
    if (desc.startsWith("D")) {
      return DSTORE;
    }
    if (desc.startsWith("J")) {
      return LSTORE;
    }
    throw new IllegalStateException("Unknown description " + desc);
  }

  public static int getLoad(final String desc) {
    if (desc.startsWith("[") || desc.startsWith("L")) {
      return ALOAD;
    }
    if (desc.startsWith("I") || desc.startsWith("S") || desc.startsWith("Z")
        || desc.startsWith("B") || desc.startsWith("C")) {
      return ILOAD;
    }
    if (desc.startsWith("F")) {
      return FLOAD;
    }
    if (desc.startsWith("D")) {
      return DLOAD;
    }
    if (desc.startsWith("J")) {
      return LLOAD;
    }
    throw new IllegalStateException("Unknown description " + desc);
  }

  public static boolean isInstrumentationClass(final String name) {
    return name.contains("LeakDetector");
  }

  public static boolean isSdk(final String name) {
    return name.startsWith("java/") || name.startsWith("sun/")
        || name.startsWith("com/apple/");
  }

  public static byte[] writeBytecode(final ClassNode cn) {
    final ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
    cn.accept(cw);
    return cw.toByteArray();
  }

  public static void saveToFile(final String className, final byte[] bytecode) {
    final String name = className.substring(className.lastIndexOf('/') + 1);
    try {
      final FileOutputStream fos = new FileOutputStream(name + ".class");
      fos.write(bytecode);
      fos.close();
    } catch (final IOException e) {
      Logger.getAnonymousLogger().log(Level.SEVERE,
          "error writing class " + name, e);
    }
  }

  public static byte[] removeJsrRets(final String className,
      final byte[] bytecode) {
    final ClassWriter writer = new ClassWriter(0);
    new ClassReader(bytecode).accept(new ClassVisitor(ASM4, writer) {

      @Override
      public MethodVisitor visitMethod(final int access, final String name,
          final String desc, final String signature, final String[] exceptions) {
        return new JSRInlinerAdapter(super.visitMethod(access, name, desc,
            signature, exceptions), access, className, desc, signature,
            exceptions);
      }
    }, ASM4);
    return writer.toByteArray();
  }
}
